searchData={"items":[{"type":"module","title":"Reactor","doc":"Reactor is a dynamic, concurrent, dependency resolving saga orchestrator.","ref":"Reactor.html"},{"type":"module","title":"Usage - Reactor","doc":"You can construct a reactor using the `Reactor` Spark DSL:\n\n```elixir\ndefmodule HelloWorldReactor do\n  @moduledoc false\n  use Reactor\n\n  input :whom\n\n  step :greet, Greeter do\n    argument :whom, input(:whom)\n  end\n\n  return :greet\nend\n```\n\n    iex> Reactor.run(HelloWorldReactor, %{whom: \"Dear Reader\"})\n    {:ok, \"Hello, Dear Reader!\"}\n\nor you can build it programmatically:\n\n    iex> reactor = Builder.new()\n    ...> {:ok, reactor} = Builder.add_input(reactor, :whom)\n    ...> {:ok, reactor} = Builder.add_step(reactor, :greet, Greeter, whom: {:input, :whom})\n    ...> {:ok, reactor} = Builder.return(reactor, :greet)\n    ...> Reactor.run(reactor, %{whom: nil})\n    {:ok, \"Hello, World!\"}","ref":"Reactor.html#module-usage"},{"type":"module","title":"DSL Documentation - Reactor","doc":"#","ref":"Reactor.html#module-dsl-documentation"},{"type":"module","title":"Index - Reactor","doc":"* [reactor](#module-reactor)\n   * around\n     * argument\n     * wait_for\n   * collect\n     * argument\n     * wait_for\n   * compose\n     * argument\n     * wait_for\n   * debug\n     * argument\n     * wait_for\n   * group\n     * argument\n     * wait_for\n   * input\n   * step\n     * argument\n     * wait_for\n   * switch\n     * matches?\n\n     * default\n\n\n#","ref":"Reactor.html#module-index"},{"type":"module","title":"Docs - Reactor","doc":"","ref":"Reactor.html#module-docs"},{"type":"module","title":"reactor - Reactor","doc":"The top-level reactor DSL\n\n * [around](#module-around)\n   * argument\n   * wait_for\n * [collect](#module-collect)\n   * argument\n   * wait_for\n * [compose](#module-compose)\n   * argument\n   * wait_for\n * [debug](#module-debug)\n   * argument\n   * wait_for\n * [group](#module-group)\n   * argument\n   * wait_for\n * [input](#module-input)\n * [step](#module-step)\n   * argument\n   * wait_for\n * [switch](#module-switch)\n   * matches?\n\n   * default\n\n\n\n\n\n\n---\n\n* `:return` (`t:atom/0`) - Specify which step result to return upon completion.\n\n\n\n#","ref":"Reactor.html#module-reactor"},{"type":"module","title":"around - Reactor","doc":"Wrap a function around a group of steps.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\n\n\n* `:name` (`t:atom/0`) - Required. A unique name of the group of steps.\n\n* `:fun` - Required. The around function.  \n  See `Reactor.Step.Around` for more information.\n\n* `:allow_async?` (`t:boolean/0`) - Whether the emitted steps should be allowed to run asynchronously.  \n  Passed to the child Reactor as it's `async?` option. The default value is `false`.\n\n\n\n###","ref":"Reactor.html#module-around"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"collect - Reactor","doc":"A Reactor step which simply collects and returns it's arguments.\n\nArguments can optionally be transformed before returning.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\nExamples:\n```\ncollect :latest_release_uri do\n  argument :repository, input(:repository)\n  argument :organisation, input(:organisation)\n\n  transform fn inputs ->\n    %{uri: \"https://api.github.com/repos/#{inputs.organisation}/#{inputs.repository}/releases/latest\"}\n  end\nend\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for the step.  \n  This is used when choosing the return value of the Reactor and for arguments into other steps.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  entire argument map before it is returned. The default value is `nil`.\n\n\n\n###","ref":"Reactor.html#module-collect"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"compose - Reactor","doc":"Compose another Reactor into this one.\n\nAllows place another Reactor into this one as if it were a single step.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for the step.  \n  Allows the result of the composed reactor to be depended upon by steps\n  in this reactor.\n\n* `:reactor` - Required. The reactor module or struct to compose upon.\n\n\n\n###","ref":"Reactor.html#module-compose"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"debug - Reactor","doc":"Inserts a step which will send debug information to the `Logger`.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\nExamples:\n```\ndebug :debug do\n  argument :suss, result(:suss_step)\nend\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. A unique identifier for the step.\n\n* `:level` - The log level to send the debug information to.\n   Valid values are :emergency, :alert, :critical, :error, :warning, :notice, :info, :debug The default value is `:debug`.\n\n\n\n###","ref":"Reactor.html#module-debug"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"group - Reactor","doc":"Call functions before and after a group of steps.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for the group of steps.\n\n* `:before_all` - Required. The before function.  \n  See `Reactor.Step.Group` for more information.\n\n* `:after_all` - Required. The after function.  \n  See `Reactor.Step.Group` for more information.\n\n* `:allow_async?` (`t:boolean/0`) - Whether the emitted steps should be allowed to run asynchronously.  \n  Passed to the child Reactor as it's `async?` option. The default value is `true`.\n\n\n\n###","ref":"Reactor.html#module-group"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"input - Reactor","doc":"Specifies an input to the Reactor.\n\nAn input is a value passed in to the Reactor when executing.\nIf a Reactor were a function, these would be it's arguments.\n\nInputs can be transformed with an arbitrary function before being passed\nto any steps.\n\n\n\n\nExamples:\n```\ninput :name\n\n```\n```\ninput :age do\n  transform &String.to_integer/1\nend\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for this input.  \n  The name is used to allow steps to depend on it.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  input before it is passed to any steps. The default value is `nil`.\n\n\n\n\n\n\n#","ref":"Reactor.html#module-input"},{"type":"module","title":"step - Reactor","doc":"Specifies a Reactor step.\n\nSteps are the unit of work in a Reactor.  Reactor will calculate the\ndependencies graph between the steps and execute as many as it can in each\niteration.\n\nSee the `Reactor.Step` behaviour for more information.\n\n\n * [argument](#module-argument)\n * [wait_for](#module-wait_for)\n\nExamples:\n```\nstep :create_user, MyApp.Steps.CreateUser do\n  argument :username, input(:username)\n  argument :password_hash, result(:hash_password)\nend\n\n```\n```\nstep :hash_password do\n  argument :password, input(:password)\n\n  run fn %{password: password}, _ ->\n    {:ok, Bcrypt.hash_pwd_salt(password)}\n  end\nend\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for the step.  \n  This is used when choosing the return value of the Reactor and for arguments into\n  other steps.\n\n* `:impl` - The step implementation.  \n  Provides an implementation for the step with the named module.  The\n  module must implement the `Reactor.Step` behaviour.\n\n* `:run` - Provide an anonymous function which implements the `run/3` callback.  \n  You cannot provide this option at the same time as the `impl` argument.\n\n* `:undo` - Provide an anonymous function which implements the `undo/4` callback.  \n  You cannot provide this option at the same time as the `impl` argument.\n\n* `:compensate` - Provide an anonymous function which implements the `undo/4` callback.  \n  You cannot provide this option at the same time as the `impl` argument.\n\n* `:max_retries` - The maximum number of times that the step can be retried before failing.  \n  This is only used when the result of the `compensate/4` callback is\n  `:retry`. The default value is `:infinity`.\n\n* `:async?` (`t:boolean/0`) - When set to true the step will be executed asynchronously via Reactor's\n  `TaskSupervisor`. The default value is `true`.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  entire argument map before it is passed to the step. The default value is `nil`.\n\n\n\n###","ref":"Reactor.html#module-step"},{"type":"module","title":"argument - Reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.\n\n\n\n\nExamples:\n```\nargument :name, input(:name)\n\n```\n```\nargument :year, input(:date, [:year])\n\n```\n```\nargument :user, result(:create_user)\n\n```\n```\nargument :user_id, result(:create_user) do\n  transform & &1.id\nend\n\n```\n```\nargument :user_id, result(:create_user, [:id])\n\n```\n```\nargument :three, value(3)\n\n```\n\n\n* `:name` (`t:atom/0`) - Required. The name of the argument which will be used as the key in the\n  `arguments` map passed to the implementation.\n\n* `:source` - Required. What to use as the source of the argument.  \n  See `Reactor.Dsl.Argument` for more information.\n\n* `:transform` - An optional transformation function which can be used to modify the\n  argument before it is passed to the step. The default value is `nil`.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-argument"},{"type":"module","title":"wait_for - Reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`\n\n\n\n\nExamples:\n```\nwait_for :create_user\n```\n\n\n* `:names` - Required. The name of the step to wait for.\n\n\n\n\n\n\n\n#","ref":"Reactor.html#module-wait_for"},{"type":"module","title":"switch - Reactor","doc":"Use a predicate to determine which steps should be executed.\n\n\n * [matches?](#module-matches?)\n\n * [default](#module-default)\n\n\n\n\n* `:name` (`t:atom/0`) - Required. A unique name for the switch.\n\n* `:allow_async?` (`t:boolean/0`) - Whether the emitted steps should be allowed to run asynchronously. The default value is `true`.\n\n* `:on` - Required. The value to match against.\n\n\n\n###","ref":"Reactor.html#module-switch"},{"type":"module","title":"matches? - Reactor","doc":"A group of steps to run when the predicate matches.\n\n\n\n\n\n\n* `:predicate` - Required. A one-arity function which is used to match the switch input.  \n  If the switch returns a truthy value, then the nested steps will be run.\n\n* `:allow_async?` (`t:boolean/0`) - Whether the emitted steps should be allowed to run asynchronously. The default value is `true`.\n\n* `:return` (`t:atom/0`) - Specify which step result to return upon completion.\n\n\n\n\n\n\n###","ref":"Reactor.html#module-matches"},{"type":"module","title":"default - Reactor","doc":"If none of the `matches?` branches match the input, then the `default`\nsteps will be run if provided.\n\n\n\n\n\n\n* `:return` (`t:atom/0`) - Specify which step result to return upon completion.","ref":"Reactor.html#module-default"},{"type":"function","title":"Reactor.run/4","doc":"Run a reactor.","ref":"Reactor.html#run/4"},{"type":"type","title":"Reactor.async_option/0","doc":"When set to `false` forces the Reactor to run every step synchronously,\nregardless of the step configuration.\n\nDefaults to `true`.","ref":"Reactor.html#t:async_option/0"},{"type":"type","title":"Reactor.concurrency_key_option/0","doc":"Use a `Reactor.Executor.ConcurrencyTracker.pool_key` to allow this Reactor to\nshare it's concurrency pool with other Reactor instances.\n\nIf you do not specify one then the Reactor will initialise a new pool and\nplace it in it's context for any child Reactors to re-use.\n\nOnly used if `async?` is set to `true`.","ref":"Reactor.html#t:concurrency_key_option/0"},{"type":"type","title":"Reactor.context/0","doc":"","ref":"Reactor.html#t:context/0"},{"type":"type","title":"Reactor.context_arg/0","doc":"","ref":"Reactor.html#t:context_arg/0"},{"type":"type","title":"Reactor.halt_timeout_option/0","doc":"How long to wait for asynchronous steps to complete when halting.\n\nDefaults to 5000ms.","ref":"Reactor.html#t:halt_timeout_option/0"},{"type":"type","title":"Reactor.inputs/0","doc":"","ref":"Reactor.html#t:inputs/0"},{"type":"type","title":"Reactor.max_concurrency_option/0","doc":"Specify the maximum number of asynchronous steps which can be run in parallel.\n\nDefaults to the result of `System.schedulers_online/0`.  Only used if\n`async?` is set to `true`.","ref":"Reactor.html#t:max_concurrency_option/0"},{"type":"type","title":"Reactor.max_iterations_option/0","doc":"The maximum number of iterations which after which the Reactor will halt.\n\nDefaults to `:infinity`.","ref":"Reactor.html#t:max_iterations_option/0"},{"type":"type","title":"Reactor.options/0","doc":"","ref":"Reactor.html#t:options/0"},{"type":"type","title":"Reactor.state/0","doc":"","ref":"Reactor.html#t:state/0"},{"type":"type","title":"Reactor.t/0","doc":"","ref":"Reactor.html#t:t/0"},{"type":"type","title":"Reactor.timeout_option/0","doc":"Specify the amount of execution time after which to halt processing.\n\nNote that this is not a hard limit. The Reactor will stop when the first step\ncompletes _after_ the timeout has expired.\n\nDefaults to `:infinity`.","ref":"Reactor.html#t:timeout_option/0"},{"type":"behaviour","title":"Reactor.Step","doc":"The Step behaviour and struct.\n\nImplement this behaviour to make steps for your Reactor.","ref":"Reactor.Step.html"},{"type":"function","title":"Reactor.Step.can?/2","doc":"Find out of a step has a capability.","ref":"Reactor.Step.html#can?/2"},{"type":"function","title":"Reactor.Step.compensate/4","doc":"Compensate a step","ref":"Reactor.Step.html#compensate/4"},{"type":"callback","title":"Reactor.Step.compensate/4","doc":"Compensate for the failure of the step.\n\n> Do not implement this callback if your step doesn't support compensation.\n\nIf `run/3` returned an error then this callback will be called the error\nreason and the original arguments.\n\nThis provides you the opportunity to handle the error in a number of ways and\ndirect the reactor as to what to do next.","ref":"Reactor.Step.html#c:compensate/4"},{"type":"callback","title":"Arguments - Reactor.Step.compensate/4","doc":"- `reason` - the error reason returned from `c:run/3`.\n  - `arguments` - the arguments passed to the step.\n  - `context` - the reactor context.\n  - `options` - a keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:compensate/4-arguments"},{"type":"callback","title":"Return values - Reactor.Step.compensate/4","doc":"- `{:continue, value}` if you're able to provide a valid result for the step\n    (perhaps by re-running the original computation) then return that within a\n    `:continue` tuple and execution will continue as planned.\n  - `:ok` the step was successfully compensated and the reactor should\n    continue undoing upstream changes.\n  - `:retry` or `{:retry, reason}` if you would like the reactor to attempt to\n    re-run the step. You can optionally supply an error reason which will be\n    used in the event that the step runs out of retries, otherwise a\n    `Reactor.Error.RetriesExceededError` will be used.\n  - `{:error, reason}` if compensation was unsuccessful.","ref":"Reactor.Step.html#c:compensate/4-return-values"},{"type":"function","title":"Reactor.Step.run/3","doc":"Execute a step.","ref":"Reactor.Step.html#run/3"},{"type":"callback","title":"Reactor.Step.run/3","doc":"Execute the step.\n\nThis is the function that implements the behaviour you wish to execute.  You\nwill receive arguments as per the `t:Step.t` definition along with their\ncorresponding values as a map and a copy of the current reactor context.","ref":"Reactor.Step.html#c:run/3"},{"type":"callback","title":"Arguments - Reactor.Step.run/3","doc":"- `arguments` - A map of arguments as per the `t:Step.t` definition we're\n    called from.\n  - `context` - The reactor context.\n  - `options` - A keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:run/3-arguments"},{"type":"callback","title":"Return values - Reactor.Step.run/3","doc":"- `{:ok, value}` the step completed successfully it returns the value in an\n    ok tuple.\n  - `{:ok, value, [step]}` the step completed successfully and wants to add\n    new steps to the reactor.\n  - `{:error, reason}` the if step failed, return an error tuple.\n  - `:retry` or `{:retry, reason}` the step failed, but is retryable.  You can\n    optionally supply an error reason which will be used in the event that the\n    step runs out of retries, otherwise a `Reactor.Error.RetriesExceededError`\n    will be used.\n  - `{:halt, reason}` terminate (or pause) reactor execution.  If there are\n    actively running steps the reactor will wait for them to finish and then\n    return the incomplete state for later resumption.","ref":"Reactor.Step.html#c:run/3-return-values"},{"type":"function","title":"Reactor.Step.undo/4","doc":"Undo a step","ref":"Reactor.Step.html#undo/4"},{"type":"callback","title":"Reactor.Step.undo/4","doc":"Undo a previously successful execution of the step.\n\n> Do not implement this callback if your step doesn't support undoing.\n\nThis callback is called when the reactor encounters an unhandled error later\nin it's execution run and must undo the work previously done.","ref":"Reactor.Step.html#c:undo/4"},{"type":"callback","title":"Arguments - Reactor.Step.undo/4","doc":"- `value` - the return value of the previously successful call to `c:run/3`.\n  - `arguments` - the arguments passed to the step.\n  - `context` - the reactor context.\n  - `options` - a keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:undo/4-arguments"},{"type":"callback","title":"Return values - Reactor.Step.undo/4","doc":"- `:ok` the step was successfully undo and the reactor should continue\n    rolling back.\n  - `{:error, reason}` there was an error while attempting to compensate.  The\n    reactor will collect the error and continue rolling back.\n  - `:retry` if you would like the reactor to attempt to undo the again later\n    - possibly in the case of a network failure for example.","ref":"Reactor.Step.html#c:undo/4-return-values"},{"type":"type","title":"Reactor.Step.capability/0","doc":"Optional capabilities which may be implemented by the step module.\n\nThis allows us to optimise out calls steps which cannot be undone, etc.","ref":"Reactor.Step.html#t:capability/0"},{"type":"type","title":"Reactor.Step.compensate_result/0","doc":"Possible valid return values for the `c:compensate/4` callback.","ref":"Reactor.Step.html#t:compensate_result/0"},{"type":"type","title":"Reactor.Step.run_result/0","doc":"Possible valid return values for the `c:run/3` callback.","ref":"Reactor.Step.html#t:run_result/0"},{"type":"type","title":"Reactor.Step.step/0","doc":"","ref":"Reactor.Step.html#t:step/0"},{"type":"type","title":"Reactor.Step.t/0","doc":"","ref":"Reactor.Step.html#t:t/0"},{"type":"type","title":"Reactor.Step.undo_result/0","doc":"Possible valid return values for the `c:undo/4` callback.","ref":"Reactor.Step.html#t:undo_result/0"},{"type":"module","title":"Reactor.Step.AnonFn","doc":"The built-in step for executing in-line DSL anonymous functions.\n\nThis step assumes that it is being called as per the\n`:spark_function_behaviour` semantics.","ref":"Reactor.Step.AnonFn.html"},{"type":"module","title":"Reactor.Step.Around","doc":"Wrap the execution of a number of steps in a function.\n\nThis allows you to provide custom context and filter the provided steps as\nneeded.","ref":"Reactor.Step.Around.html"},{"type":"module","title":"Options - Reactor.Step.Around","doc":"* `fun` - a four-arity function that will be called when executing this step.\n* `steps` - a list of steps which are will be provided to the above mentioned\n  function.\n* `allow_async?` - a boolean indicating whether the nested steps can be\n  executed asynchronously or must remain within the current process.","ref":"Reactor.Step.Around.html#module-options"},{"type":"module","title":"Wrapper function - Reactor.Step.Around","doc":"Your around function will be called by this step and will be passed the\nfollowing arguments:\n\n* `arguments` - the arguments passed to the step.\n* `context` - the context passed to the step.\n* `steps` - the list of steps passed in the options.\n* `callback` - a 3 arity function that you can call to execute steps.\n\nThis provides you the opportunity to modify the arguments, context and list of\nsteps to be executed.  You then can call the callback with the modified\narguments, context and steps and they will be executed in a Reactor of their\nown.  The callback will return `{:ok, results}` where results is a map of all\nof the step results by name, or an error tuple.\n\nYou can then modify the result in any way before returning it as the return of\nthe around step.","ref":"Reactor.Step.Around.html#module-wrapper-function"},{"type":"module","title":"Callback function - Reactor.Step.Around","doc":"The callback function will spawn a separate Reactor and run provided steps to\ncompletion using `arguments` as input.\n\nIt expects the following three arguments to be passed:\n\n1. `arguments` - a map of arguments to be used as input to the nested Reactor.\n2. `context` - the context passed to the nested Reactor.\n3. `steps` - the list of steps which will be executed in the nested Reactor.","ref":"Reactor.Step.Around.html#module-callback-function"},{"type":"module","title":"Example - Reactor.Step.Around","doc":"You could use a function like that below to cause some steps to be executed\ninside an Ecto database transaction.\n\n```elixir\ndef in_transaction(arguments, context, steps, callback) do\n  MyApp.Repo.transaction(fn ->\n    case callback.(arguments, context, steps) do\n      {:ok, results} -> result\n      {:error, reason} -> raise reason\n    end\n  end)\nend\n```","ref":"Reactor.Step.Around.html#module-example"},{"type":"function","title":"Reactor.Step.Around.around/4","doc":"","ref":"Reactor.Step.Around.html#around/4"},{"type":"type","title":"Reactor.Step.Around.allow_async_option/0","doc":"Should the inner Reactor be allowed to run tasks asynchronously?\n\nOptional. Defaults to `true`.","ref":"Reactor.Step.Around.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Around.around_fun/0","doc":"The type signature for the \"around\" function.","ref":"Reactor.Step.Around.html#t:around_fun/0"},{"type":"type","title":"Reactor.Step.Around.callback/0","doc":"The type signature for the provided callback function.","ref":"Reactor.Step.Around.html#t:callback/0"},{"type":"type","title":"Reactor.Step.Around.function_option/0","doc":"The MFA or 4-arity function which this step will call.","ref":"Reactor.Step.Around.html#t:function_option/0"},{"type":"type","title":"Reactor.Step.Around.options/0","doc":"","ref":"Reactor.Step.Around.html#t:options/0"},{"type":"type","title":"Reactor.Step.Around.steps_option/0","doc":"The initial steps to pass into the \"around\" function.\n\nOptional.","ref":"Reactor.Step.Around.html#t:steps_option/0"},{"type":"module","title":"Reactor.Step.Compose","doc":"A built-in step which can embed one reactor inside another.\n\nThis is different to the `Builder.compose` and DSL `compose` methods.  Those\nmethods build a new reactor by combining the steps of the two input reactors,\nwhereas this step expands the provided reactor at runtime and dynamically\ninserts it's steps into the running reactor.\n\nIf emitting the reactor's steps into the current reactor would be recursive,\nthen the reactor is directly executed within the step using `Reactor.run/4`.","ref":"Reactor.Step.Compose.html"},{"type":"module","title":"Reactor.Step.ComposeWrapper","doc":"When doing run-time composition of Reactors we need to dynamically rewrite any\ndynamically emitted steps to have the correct namespace.\n\nYes, this gets hairy, fast.\n\nThis is dynamically injected into steps by `Reactor.Step.Compose` - you\nprobably don't want to use this unless you're sure what you're doing.","ref":"Reactor.Step.ComposeWrapper.html"},{"type":"module","title":"Options - Reactor.Step.ComposeWrapper","doc":"* `original` - the original value of the Step's `impl` key.\n  * `prefix` - a list of values to be placed in the `name` before the original value.","ref":"Reactor.Step.ComposeWrapper.html#module-options"},{"type":"module","title":"Reactor.Step.Debug","doc":"A very simple step which sends debug information to `Logger`.","ref":"Reactor.Step.Debug.html"},{"type":"module","title":"Options - Reactor.Step.Debug","doc":"* `level` - the log level to use.","ref":"Reactor.Step.Debug.html#module-options"},{"type":"type","title":"Reactor.Step.Debug.level_option/0","doc":"","ref":"Reactor.Step.Debug.html#t:level_option/0"},{"type":"type","title":"Reactor.Step.Debug.options/0","doc":"","ref":"Reactor.Step.Debug.html#t:options/0"},{"type":"module","title":"Reactor.Step.Group","doc":"Wrap the execution of a number of steps with before/after functions.\n\nUnlike `Reactor.Step.Around`, this step doesn't need to run a nested Reactor\ninstance, but instead can emit the steps directly into the parent Reactor.","ref":"Reactor.Step.Group.html"},{"type":"module","title":"Options - Reactor.Step.Group","doc":"* `before` - a three-arity function that will be called before running any\n  child steps.\n* `after` - a three-arity function that will be called after running any\n  emitted steps.\n* `allow_async?` - a boolean indicating whether the emitted steps can be\n  executed asynchronously or must remain within the current process.","ref":"Reactor.Step.Group.html#module-options"},{"type":"module","title":"Before function - Reactor.Step.Group","doc":"The before function will be passed the following arguments:\n\n1. `arguments` - the values of any step arguments needed by the group.\n2. `context` - the Reactor context.\n3. `steps` - the list of steps passed in the options.\n\nThis provides you the opportunity to modify the arguments, context and list of\nsteps to be executed.\n\nThe successful return value should be `{:ok, arguments, context, steps}`.  The\nreturned arguments will be used to provide any `input` arguments to nested\nsteps.\n\n#","ref":"Reactor.Step.Group.html#module-before-function"},{"type":"module","title":"Example - Reactor.Step.Group","doc":"```elixir\ndef no_time_travel(arguments, context, steps) do\n  steps = steps\n    |> Enum.filter(&(&1.name == :program_time_circuits))\n\n  arguments = arguments\n    |> Map.delete(:destination_time)\n\n  {:ok, arguments, context, steps}\nend\n```","ref":"Reactor.Step.Group.html#module-example"},{"type":"module","title":"After function - Reactor.Step.Group","doc":"The after function will be called with a single argument; a map of the nested\nstep results.\n\nThe successful return value should be `{:ok, any}` where `any` will be treated\nas the result of the group.\n\n```elixir\ndef find_current_year(results) do\n  case Map.fetch(results, :time_circuits) do\n    {:ok, %{present_time: present_time}} -> {:ok, present_time.year}\n    _ -> {:error, \"Unable to read the present time from the time circuits\"}\n  end\nend\n```","ref":"Reactor.Step.Group.html#module-after-function"},{"type":"type","title":"Reactor.Step.Group.after_fun/0","doc":"The after function.","ref":"Reactor.Step.Group.html#t:after_fun/0"},{"type":"type","title":"Reactor.Step.Group.after_option/0","doc":"The MFA or 1-arity function which this step will call after successfully\nrunning the steps.","ref":"Reactor.Step.Group.html#t:after_option/0"},{"type":"type","title":"Reactor.Step.Group.allow_async_option/0","doc":"Should the emitted steps be allowed to run asynchronously?\n\nOptional. Defaults to `true`.","ref":"Reactor.Step.Group.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Group.before_fun/0","doc":"The before function.","ref":"Reactor.Step.Group.html#t:before_fun/0"},{"type":"type","title":"Reactor.Step.Group.before_option/0","doc":"The MFA or 3-arity function which this step will call before running any\nsteps.","ref":"Reactor.Step.Group.html#t:before_option/0"},{"type":"type","title":"Reactor.Step.Group.options/0","doc":"","ref":"Reactor.Step.Group.html#t:options/0"},{"type":"type","title":"Reactor.Step.Group.steps_option/0","doc":"The initial steps to pass into the \"before\" function.\n\nOptional.","ref":"Reactor.Step.Group.html#t:steps_option/0"},{"type":"module","title":"Reactor.Step.ReturnAllArguments","doc":"A very simple step which simply returns all it's arguments unchanged.","ref":"Reactor.Step.ReturnAllArguments.html"},{"type":"module","title":"Reactor.Step.ReturnArgument","doc":"A very simple step which simply returns the named argument, if provided.","ref":"Reactor.Step.ReturnArgument.html"},{"type":"module","title":"Options. - Reactor.Step.ReturnArgument","doc":"* `argument` - the name of the argument to return.","ref":"Reactor.Step.ReturnArgument.html#module-options"},{"type":"module","title":"Reactor.Step.Switch","doc":"Conditionally decide which steps should be run at runtime.","ref":"Reactor.Step.Switch.html"},{"type":"module","title":"Options - Reactor.Step.Switch","doc":"* `matches` - a list of match consisting of predicates and a list of steps to\n  execute if the predicate returns a truthy value.  See `t:matches` for more\n  information.  Required.\n* `default` - a list of steps to execute if none of the predicates match.\n  Optional.\n* `allow_async?` - a boolean indicating whether to allow the steps to be\n  executed asynchronously.  Optional.  Defaults to `true`.\n* `on` - the name of the argument to pass into the predicates.  If this\n  argument is not provided to this step, then an error will be returned.","ref":"Reactor.Step.Switch.html#module-options"},{"type":"module","title":"Branching behaviour - Reactor.Step.Switch","doc":"Each of the predicates in `matches` are tried in order, until either one\nreturns a truthy value, or all the matches are exhausted.\n\nIf there is a match, then the matching steps are emitted into the parent\nrunning Reactor.\n\nIn the case that no match is found, then the steps provided in the `default`\noption are emitted.  If no default is provided, then an error is returned.\n\n> ##","ref":"Reactor.Step.Switch.html#module-branching-behaviour"},{"type":"module","title":"Tip {: .tip} - Reactor.Step.Switch","doc":">\n> Execution of predicates stops once the first match is found.  This means\n> that if multiple predicates potentially match, the subsequent ones will\n> never be called.","ref":"Reactor.Step.Switch.html#module-tip-tip"},{"type":"module","title":"Returning - Reactor.Step.Switch","doc":"By default the step returns `nil` as it's result.\n\nYou can have the step return the result of a branch by adding a step to the\nbranch with the same name as the switch which returns the expected value.\nThis will be handled by normal Reactor step emission rules.","ref":"Reactor.Step.Switch.html#module-returning"},{"type":"type","title":"Reactor.Step.Switch.allow_async_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Switch.default_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:default_option/0"},{"type":"type","title":"Reactor.Step.Switch.match_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:match_option/0"},{"type":"type","title":"Reactor.Step.Switch.matches/0","doc":"A list of predicates and steps to execute if the predicate returns a truthy\nvalue.","ref":"Reactor.Step.Switch.html#t:matches/0"},{"type":"type","title":"Reactor.Step.Switch.on_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:on_option/0"},{"type":"type","title":"Reactor.Step.Switch.options/0","doc":"","ref":"Reactor.Step.Switch.html#t:options/0"},{"type":"type","title":"Reactor.Step.Switch.predicate/0","doc":"A predicate is a 1-arity function.  It can return anything.  Any result which\nis not `nil` or `false` is considered true.","ref":"Reactor.Step.Switch.html#t:predicate/0"},{"type":"module","title":"Reactor.Step.Transform","doc":"The built-in step for executing input and argument transformations.\n\nThis step assumes that it is being executed as per `:spark_function_behaviour`\nsemantics.","ref":"Reactor.Step.Transform.html"},{"type":"module","title":"Reactor.Step.TransformAll","doc":"A built-in step which applies a transformation function to all it's arguments.\n\nThe returned map is used as the arguments to the step, instead of the step's\ndefined arguments.","ref":"Reactor.Step.TransformAll.html"},{"type":"module","title":"Reactor.Argument","doc":"A step argument.","ref":"Reactor.Argument.html"},{"type":"function","title":"Reactor.Argument.from_input/3","doc":"Build an argument which refers to a reactor input with an optional\ntransformation applied.","ref":"Reactor.Argument.html#from_input/3"},{"type":"function","title":"Example - Reactor.Argument.from_input/3","doc":"iex> Argument.from_input(:argument_name, :input_name, &String.to_integer/1)","ref":"Reactor.Argument.html#from_input/3-example"},{"type":"function","title":"Reactor.Argument.from_result/3","doc":"Build an argument which refers to the result of another step with an optional\ntransformation applied.","ref":"Reactor.Argument.html#from_result/3"},{"type":"function","title":"Example - Reactor.Argument.from_result/3","doc":"iex> Argument.from_result(:argument_name, :step_name, &Atom.to_string/1)","ref":"Reactor.Argument.html#from_result/3-example"},{"type":"function","title":"Reactor.Argument.from_value/3","doc":"Build an argument which refers to a statically defined value.","ref":"Reactor.Argument.html#from_value/3"},{"type":"function","title":"Example - Reactor.Argument.from_value/3","doc":"iex> Argument.from_value(:argument_name, 10)","ref":"Reactor.Argument.html#from_value/3-example"},{"type":"macro","title":"Reactor.Argument.has_sub_path/1","doc":"Validate that the argument source has a sub_path","ref":"Reactor.Argument.html#has_sub_path/1"},{"type":"macro","title":"Reactor.Argument.has_transform/1","doc":"Validate that the argument has a transform.","ref":"Reactor.Argument.html#has_transform/1"},{"type":"macro","title":"Reactor.Argument.is_argument/1","doc":"Validate that the argument is an Argument struct.","ref":"Reactor.Argument.html#is_argument/1"},{"type":"macro","title":"Reactor.Argument.is_from_input/1","doc":"Validate that the argument refers to a reactor input.","ref":"Reactor.Argument.html#is_from_input/1"},{"type":"macro","title":"Reactor.Argument.is_from_result/1","doc":"Validate that the argument refers to a step result.","ref":"Reactor.Argument.html#is_from_result/1"},{"type":"macro","title":"Reactor.Argument.is_from_value/1","doc":"Validate that the argument contains a static value.","ref":"Reactor.Argument.html#is_from_value/1"},{"type":"type","title":"Reactor.Argument.t/0","doc":"","ref":"Reactor.Argument.html#t:t/0"},{"type":"protocol","title":"Reactor.Argument.Build","doc":"A protocol which can be used to convert something into an argument.","ref":"Reactor.Argument.Build.html"},{"type":"function","title":"Reactor.Argument.Build.build/1","doc":"Convert the input into an argument.","ref":"Reactor.Argument.Build.html#build/1"},{"type":"type","title":"Reactor.Argument.Build.t/0","doc":"All the types that implement this protocol.","ref":"Reactor.Argument.Build.html#t:t/0"},{"type":"module","title":"Reactor.Builder","doc":"Build a new Reactor programmatically.\n\nYou don't _have_ to use the Reactor DSL to create a Reactor.  The functions in\nthis module allow you to define a Reactor programmatically.  This is\nespecially useful if you need to create a reactor dynamically (maybe based on\na UI such as [React Flow](https://reactflow.dev/)).","ref":"Reactor.Builder.html"},{"type":"module","title":"Example - Reactor.Builder","doc":"```elixir\nreactor = Builder.new()\n{:ok, reactor} = Builder.add_input(reactor, :name)\nargument = Argument.from_input(:name)\n{:ok, reactor} = Builder.add_step(reactor, :greet, [argument])\n{:ok, reactor} = Builder.return(reactor, :greet)\n```","ref":"Reactor.Builder.html#module-example"},{"type":"function","title":"Reactor.Builder.add_input/3","doc":"Add a named input to the Reactor.\n\nThis both places the input in the Reactor for later input validation and adds\nsteps to the Reactor which will emit and (possibly) transform the input.","ref":"Reactor.Builder.html#add_input/3"},{"type":"function","title":"Reactor.Builder.add_input!/3","doc":"Raising version of `add_input/2..3`.","ref":"Reactor.Builder.html#add_input!/3"},{"type":"function","title":"Reactor.Builder.add_step/5","doc":"Add a step to the Reactor.\n\nAdd a new step to the Reactor.  Rewrites input arguments to use the result of\nthe input steps and injects transformation steps as required.","ref":"Reactor.Builder.html#add_step/5"},{"type":"function","title":"Reactor.Builder.add_step!/5","doc":"Raising version of `add_step/3..5`.","ref":"Reactor.Builder.html#add_step!/5"},{"type":"function","title":"Reactor.Builder.compose/4","doc":"Compose another Reactor inside this one.\n\nWhenever possible this function will extract the steps from inner Reactor and\nplace them inside the parent Reactor.  In order to achieve this the composer\nwill rename the steps to ensure that there are no conflicts.\n\nIf you're attempting to create a recursive Reactor (ie compose a Reactor\nwithin itself) then this will be detected and runtime composition will be used\ninstead.  See `Reactor.Step.Compose` for more details.","ref":"Reactor.Builder.html#compose/4"},{"type":"function","title":"Reactor.Builder.compose!/4","doc":"Raising version of `compose/4`.","ref":"Reactor.Builder.html#compose!/4"},{"type":"function","title":"Reactor.Builder.new/1","doc":"Build a new, empty Reactor.\n\nOptionally an identifier for the Reactor. This is primarily used for recursive\ncomposition tracking.","ref":"Reactor.Builder.html#new/1"},{"type":"function","title":"Reactor.Builder.new_step/4","doc":"Build a step which can be added to a reactor at runtime.\n\nNote that the built step doesn't support transformations - you should add an\nadditional step to do the transformation needed (this is what `add_step/5`\ndoes anyway).","ref":"Reactor.Builder.html#new_step/4"},{"type":"function","title":"Reactor.Builder.new_step!/4","doc":"Raising version of `new_step/2..4`.","ref":"Reactor.Builder.html#new_step!/4"},{"type":"function","title":"Reactor.Builder.return/2","doc":"Specify the return value of the Reactor.\n\nThe return value must be the name of a step.","ref":"Reactor.Builder.html#return/2"},{"type":"function","title":"Reactor.Builder.return!/2","doc":"Raising version of `return/2`.","ref":"Reactor.Builder.html#return!/2"},{"type":"type","title":"Reactor.Builder.arguments_transform/0","doc":"Optionally transform all the arguments into new arguments","ref":"Reactor.Builder.html#t:arguments_transform/0"},{"type":"type","title":"Reactor.Builder.async?/0","doc":"Should the step be run asynchronously?","ref":"Reactor.Builder.html#t:async?/0"},{"type":"type","title":"Reactor.Builder.context/0","doc":"Optional context which will be merged with the reactor context when calling this step.","ref":"Reactor.Builder.html#t:context/0"},{"type":"type","title":"Reactor.Builder.impl/0","doc":"","ref":"Reactor.Builder.html#t:impl/0"},{"type":"type","title":"Reactor.Builder.max_retries/0","doc":"How many times is the step allowed to retry?","ref":"Reactor.Builder.html#t:max_retries/0"},{"type":"type","title":"Reactor.Builder.ref/0","doc":"","ref":"Reactor.Builder.html#t:ref/0"},{"type":"type","title":"Reactor.Builder.step_argument/0","doc":"","ref":"Reactor.Builder.html#t:step_argument/0"},{"type":"type","title":"Reactor.Builder.step_options/0","doc":"","ref":"Reactor.Builder.html#t:step_options/0"},{"type":"module","title":"Reactor.Builder.Compose","doc":"Handle composition of Reactors for the builder.\n\nThe composition logic was getting complicated enough that it seemed sensible\nto extract it from the builder - if only to aid readability.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.compose/4`.","ref":"Reactor.Builder.Compose.html"},{"type":"function","title":"Reactor.Builder.Compose.compose/4","doc":"Compose another Reactor inside this one.","ref":"Reactor.Builder.Compose.html#compose/4"},{"type":"function","title":"Reactor.Builder.Compose.do_static_compose/4","doc":"","ref":"Reactor.Builder.Compose.html#do_static_compose/4"},{"type":"module","title":"Reactor.Builder.Input","doc":"Handle adding inputs to Reactors for the builder.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.add_input/3`.","ref":"Reactor.Builder.Input.html"},{"type":"function","title":"Reactor.Builder.Input.add_input/3","doc":"Add a named input to the reactor.","ref":"Reactor.Builder.Input.html#add_input/3"},{"type":"module","title":"Reactor.Builder.Step","doc":"Handle building and adding steps to Reactors for the builder.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.new_step/4` and `Reactor.Builder.add_step/5`.","ref":"Reactor.Builder.Step.html"},{"type":"function","title":"Reactor.Builder.Step.add_step/5","doc":"Build and add a new step to a Reactor.","ref":"Reactor.Builder.Step.html#add_step/5"},{"type":"function","title":"Reactor.Builder.Step.new_step/4","doc":"Dynamically build a new step for later use.\n\nYou're most likely to use this when dynamically returning new steps from an\nexisting step.","ref":"Reactor.Builder.Step.html#new_step/4"},{"type":"module","title":"Reactor.Dsl.Argument","doc":"The struct used to store `argument` DSL entities.\n\nSee `d:Reactor.step.argument`.","ref":"Reactor.Dsl.Argument.html"},{"type":"function","title":"Reactor.Dsl.Argument.input/2","doc":"The `input` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#input/2"},{"type":"function","title":"Example - Reactor.Dsl.Argument.input/2","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  input :name\n\n  step :greet do\n    # here: --------↓↓↓↓↓\n    argument :name, input(:name)\n    run fn\n      %{name: nil}, _, _ -> {:ok, \"Hello, World!\"}\n      %{name: name}, _, _ -> {:ok, \"Hello, #{name}!\"}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#input/2-example"},{"type":"function","title":"Extracting nested values - Reactor.Dsl.Argument.input/2","doc":"You can provide a list of keys to extract from a data structure, similar to\n`Kernel.get_in/2` with the condition that the input value is either a struct\nor implements the `Access` protocol.","ref":"Reactor.Dsl.Argument.html#input/2-extracting-nested-values"},{"type":"function","title":"Reactor.Dsl.Argument.result/2","doc":"The `result` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#result/2"},{"type":"function","title":"Example - Reactor.Dsl.Argument.result/2","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  step :whom do\n    run fn ->\n      {:ok, Enum.random([\"Marty\", \"Doc\", \"Jennifer\", \"Lorraine\", \"George\", nil])}\n    end\n  end\n\n  step :greet do\n    # here: --------↓↓↓↓↓↓\n    argument :name, result(:whom)\n    run fn\n      %{name: nil}, _, _ -> {:ok, \"Hello, World!\"}\n      %{name: name}, _, _ -> {:ok, \"Hello, #{name}!\"}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#result/2-example"},{"type":"function","title":"Extracting nested values - Reactor.Dsl.Argument.result/2","doc":"You can provide a list of keys to extract from a data structure, similar to\n`Kernel.get_in/2` with the condition that the result is either a struct or\nimplements the `Access` protocol.","ref":"Reactor.Dsl.Argument.html#result/2-extracting-nested-values"},{"type":"function","title":"Reactor.Dsl.Argument.value/1","doc":"The `value` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#value/1"},{"type":"function","title":"Example - Reactor.Dsl.Argument.value/1","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  input :number\n\n  step :times_three do\n    argument :lhs, input(:number)\n    # here: -------↓↓↓↓↓\n    argument :rhs, value(3)\n\n    run fn args, _, _ ->\n      {:ok, args.lhs * args.rhs}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#value/1-example"},{"type":"type","title":"Reactor.Dsl.Argument.t/0","doc":"","ref":"Reactor.Dsl.Argument.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Around","doc":"The `around` DSL entity struct.\n\nSee `d:Reactor.around`.","ref":"Reactor.Dsl.Around.html"},{"type":"type","title":"Reactor.Dsl.Around.t/0","doc":"","ref":"Reactor.Dsl.Around.html#t:t/0"},{"type":"protocol","title":"Reactor.Dsl.Build","doc":"A protocol which DSL entities must implement.","ref":"Reactor.Dsl.Build.html"},{"type":"function","title":"Reactor.Dsl.Build.build/2","doc":"Build an entity into a Reactor.\n\nThis function is called during conversion of a DSL module into a Reactor\nstruct.  This allows extensions to specify the behaviour of how they want\nto alter the structure of the Reactor.","ref":"Reactor.Dsl.Build.html#build/2"},{"type":"function","title":"Reactor.Dsl.Build.transform/2","doc":"Perform any transformation that is needed to make the entity work in the\nsystem before building.\n\nSee `Spark.Dsl.Transformer` for more information.","ref":"Reactor.Dsl.Build.html#transform/2"},{"type":"function","title":"Reactor.Dsl.Build.verify/2","doc":"Perform any after-compilation verification that is needed to make the entity\nwork.\n\nSee `Spark.Dsl.Verifier` for more information.","ref":"Reactor.Dsl.Build.html#verify/2"},{"type":"type","title":"Reactor.Dsl.Build.t/0","doc":"All the types that implement this protocol.","ref":"Reactor.Dsl.Build.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Collect","doc":"The struct used to store collect DSL entities.\n\nSee `d:Reactor.collect`.","ref":"Reactor.Dsl.Collect.html"},{"type":"type","title":"Reactor.Dsl.Collect.t/0","doc":"","ref":"Reactor.Dsl.Collect.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Compose","doc":"The `compose` DSL entity struct.\n\nSee the `d:Reactor.compose`.","ref":"Reactor.Dsl.Compose.html"},{"type":"type","title":"Reactor.Dsl.Compose.t/0","doc":"","ref":"Reactor.Dsl.Compose.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Debug","doc":"The `debug` DSL entity struct.\n\nSee `d:Reactor.debug`.","ref":"Reactor.Dsl.Debug.html"},{"type":"type","title":"Reactor.Dsl.Debug.t/0","doc":"","ref":"Reactor.Dsl.Debug.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Group","doc":"The `group` DSL entity struct.\n\nSee `d:Reactor.group`.","ref":"Reactor.Dsl.Group.html"},{"type":"type","title":"Reactor.Dsl.Group.t/0","doc":"","ref":"Reactor.Dsl.Group.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Input","doc":"The struct used to store input DSL entities.\n\nSee `d:Reactor.input`.","ref":"Reactor.Dsl.Input.html"},{"type":"type","title":"Reactor.Dsl.Input.t/0","doc":"","ref":"Reactor.Dsl.Input.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Step","doc":"The struct used to store step DSL entities.\n\nSee `d:Reactor.step`.","ref":"Reactor.Dsl.Step.html"},{"type":"type","title":"Reactor.Dsl.Step.t/0","doc":"","ref":"Reactor.Dsl.Step.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch","doc":"The `switch` DSL entity struct.\n\nSee `d:Reactor.switch`.","ref":"Reactor.Dsl.Switch.html"},{"type":"type","title":"Reactor.Dsl.Switch.t/0","doc":"","ref":"Reactor.Dsl.Switch.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch.Default","doc":"The `default` DSL entity struct.\n\nSee `d:Reactor.switch.default`.","ref":"Reactor.Dsl.Switch.Default.html"},{"type":"type","title":"Reactor.Dsl.Switch.Default.t/0","doc":"","ref":"Reactor.Dsl.Switch.Default.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch.Match","doc":"The `matches?` DSL entity struct.\n\nSee `d:Reactor.switch.matches?`.","ref":"Reactor.Dsl.Switch.Match.html"},{"type":"type","title":"Reactor.Dsl.Switch.Match.t/0","doc":"","ref":"Reactor.Dsl.Switch.Match.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Verifier","doc":"Runs `Reactor.Dsl.Build.verify/2` for all the entities in the reactor.","ref":"Reactor.Dsl.Verifier.html"},{"type":"module","title":"Reactor.Dsl.WaitFor","doc":"The struct used to store `wait_for` DSL entities.\n\nSee `d:Reactor.step.wait_for`.","ref":"Reactor.Dsl.WaitFor.html"},{"type":"type","title":"Reactor.Dsl.WaitFor.t/0","doc":"","ref":"Reactor.Dsl.WaitFor.html#t:t/0"},{"type":"exception","title":"Reactor.Error.ComposeError","doc":"An error used when attempting to compose to Reactors together.","ref":"Reactor.Error.ComposeError.html"},{"type":"exception","title":"Reactor.Error.PlanError","doc":"An error thrown during the planning of a Reactor.","ref":"Reactor.Error.PlanError.html"},{"type":"exception","title":"Reactor.Error.RetriesExceededError","doc":"An error used when a step runs out of retry events and no other error is\nthrown.","ref":"Reactor.Error.RetriesExceededError.html"},{"type":"exception","title":"Reactor.Error.TransformError","doc":"An error which occurs when building and running transforms.","ref":"Reactor.Error.TransformError.html"},{"type":"module","title":"Reactor.Executor","doc":"The Reactor executor.\n\nThe executor handles the main loop of running a Reactor.\n\nThe algorithm is somewhat confusing, so here it is in pseudocode:\n\n1. Find any async tasks (from a previous loop) which are completed. Either\n   recurse or continue if none are found.\n2. Find any async steps in the plan which are ready to run (they have no\n   in-edges in the graph) and start as many as possible (given the constraints\n   of `max_concurrency` and the state of the concurrency pool).  Either start\n   over, or continue if none are found.\n3. Find a single synchronous step which is ready to run and execute it. If\n   there was one then recurse, otherwise continue.\n4. Check if there are no more steps left in the plan (there are zero\n   vertices).  If so, collect the return value and exit, otherwise recurse.\n\nWhenever a step is run, whether run synchronously or asynchronously, the\nfollowing happens:\n\n1. When the step is successful:\n  a. If the step is undoable (ie `Step.can?(module, :undo)?` returns `true`)\n     then the step and the result are stored in the Reactor's undo stack.\n  b. If the result is depended upon by another step (the graph has out-edges\n     for the step) _or_ the step is asking the reactor to halt then the\n     result is stored in the Reactor's intermediate results.\n  c. The step is removed from the graph (along with it's out-edges, freeing\n     up it's dependents to run).\n2. When the step is unsuccessful (returns an error tuple or raises):\n  a. If the step can be compensated then compensation is attempted up to five\n     times before giving up.\n  b. The reactor iterates it's undo stack calling undo on each step.\n3. When a step or compensation asks for a retry then the step is placed back\n   in the graph to be run again next iteration.","ref":"Reactor.Executor.html"},{"type":"function","title":"Reactor.Executor.run/4","doc":"Run a reactor.\n\nProvided a Reactor which has been planned and the correct inputs, then run\nthe Reactor until completion, halting or failure.\n\nYou probably shouldn't call this directly, but use `Reactor.run/4` instead.","ref":"Reactor.Executor.html#run/4"},{"type":"module","title":"Reactor.Executor.Async","doc":"Handle the asynchronous execution of a batch of steps, along with any\nmutations to the reactor or execution state.","ref":"Reactor.Executor.Async.html"},{"type":"function","title":"Reactor.Executor.Async.collect_remaining_tasks_for_shutdown/2","doc":"When the Reactor needs to shut down for any reason, we need to await all the\ncurrently running asynchronous steps and delete any task vertices.","ref":"Reactor.Executor.Async.html#collect_remaining_tasks_for_shutdown/2"},{"type":"function","title":"Reactor.Executor.Async.handle_completed_steps/2","doc":"Check to see if any steps are completed, and if so handle them.","ref":"Reactor.Executor.Async.html#handle_completed_steps/2"},{"type":"function","title":"Reactor.Executor.Async.start_steps/4","doc":"Start as many of the provided steps as possible.\n\nTakes into account he maximum concurrency and available work slots.","ref":"Reactor.Executor.Async.html#start_steps/4"},{"type":"module","title":"Reactor.Executor.ConcurrencyTracker","doc":"Manage shared concurrency pools for multiple Reactors.\n\nWhen running a Reactor you can pass the `concurrency_key` option, which will\ncause the Reactor to use the specified pool to ensure that the combined\nReactors never exceed the pool's available concurrency limit.\n\nThis avoids nested Reactors spawning too many workers and thrashing the\nsystem.","ref":"Reactor.Executor.ConcurrencyTracker.html"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.acquire/1","doc":"Attempt to acquire a concurrency allocation from the pool.\n\nReturns `:ok` if the allocation was successful, otherwise `:error`.","ref":"Reactor.Executor.ConcurrencyTracker.html#acquire/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.allocate_pool/1","doc":"Allocate a new concurrency pool and set the maximum limit.","ref":"Reactor.Executor.ConcurrencyTracker.html#allocate_pool/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Reactor.Executor.ConcurrencyTracker.html#child_spec/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.release/1","doc":"Release a concurrency allocation back to the pool.","ref":"Reactor.Executor.ConcurrencyTracker.html#release/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.release_pool/1","doc":"Release the concurrency pool.\n\nThis deletes the pool, however doesn't affect any processes currently using\nit.  No more resources can be acquired by users of the pool key.","ref":"Reactor.Executor.ConcurrencyTracker.html#release_pool/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.status/1","doc":"Report the available and maximum concurrency for a pool.","ref":"Reactor.Executor.ConcurrencyTracker.html#status/1"},{"type":"type","title":"Reactor.Executor.ConcurrencyTracker.pool_key/0","doc":"","ref":"Reactor.Executor.ConcurrencyTracker.html#t:pool_key/0"},{"type":"module","title":"Reactor.Executor.Init","doc":"Handle argument checking and state setup for a Reactor run.","ref":"Reactor.Executor.Init.html"},{"type":"module","title":"Reactor.Executor.State","doc":"Contains the reactor execution state.\n\nThis is run-time only information.","ref":"Reactor.Executor.State.html"},{"type":"type","title":"Reactor.Executor.State.t/0","doc":"","ref":"Reactor.Executor.State.html#t:t/0"},{"type":"module","title":"Reactor.Executor.StepRunner","doc":"Run an individual step, including compensation if possible.","ref":"Reactor.Executor.StepRunner.html"},{"type":"function","title":"Reactor.Executor.StepRunner.run/4","doc":"Collect the arguments and and run a step, with compensation if required.","ref":"Reactor.Executor.StepRunner.html#run/4"},{"type":"function","title":"Reactor.Executor.StepRunner.undo/5","doc":"Undo a step if possible.","ref":"Reactor.Executor.StepRunner.html#undo/5"},{"type":"module","title":"Reactor.Executor.Sync","doc":"Handle the synchronous execution of a single step, along with any mutations to\nthe reactor or execution state.","ref":"Reactor.Executor.Sync.html"},{"type":"function","title":"Reactor.Executor.Sync.run/3","doc":"Try and run a step synchronously.","ref":"Reactor.Executor.Sync.html#run/3"},{"type":"module","title":"Reactor.Info","doc":"Introspection for the Reactor DSL.","ref":"Reactor.Info.html"},{"type":"function","title":"Reactor.Info.reactor/1","doc":"reactor DSL entities","ref":"Reactor.Info.html#reactor/1"},{"type":"function","title":"Reactor.Info.reactor_options/1","doc":"reactor DSL options\n\nReturns a map containing the and any configured or default values.","ref":"Reactor.Info.html#reactor_options/1"},{"type":"function","title":"Reactor.Info.reactor_return/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Info.html#reactor_return/1"},{"type":"function","title":"Reactor.Info.reactor_return!/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Info.html#reactor_return!/1"},{"type":"function","title":"Reactor.Info.to_struct/1","doc":"Convert a reactor DSL module into a reactor struct.","ref":"Reactor.Info.html#to_struct/1"},{"type":"function","title":"Reactor.Info.to_struct!/1","doc":"Raising version of `to_struct/1`.","ref":"Reactor.Info.html#to_struct!/1"},{"type":"module","title":"Reactor.Planner","doc":"Build an execution plan for a Reactor.\n\nConverts any unplanned steps into vertices in a graph with directed edges\nbetween them representing their dependencies (arguments).","ref":"Reactor.Planner.html"},{"type":"function","title":"Reactor.Planner.plan/1","doc":"Build an execution plan for a Reactor.\n\nBuilds a graph of the step dependencies, resolves them and then builds an execution plan.","ref":"Reactor.Planner.html#plan/1"},{"type":"function","title":"Reactor.Planner.plan!/1","doc":"Raising version of `plan/1`.","ref":"Reactor.Planner.html#plan!/1"},{"type":"module","title":"Reactor.Template","doc":"Templates used to refer to some sort of computed value.","ref":"Reactor.Template.html"},{"type":"function","title":"Reactor.Template.type/0","doc":"The type for use in option schemas","ref":"Reactor.Template.html#type/0"},{"type":"type","title":"Reactor.Template.t/0","doc":"","ref":"Reactor.Template.html#t:t/0"},{"type":"module","title":"Reactor.Template.Input","doc":"The `input` template.","ref":"Reactor.Template.Input.html"},{"type":"type","title":"Reactor.Template.Input.t/0","doc":"","ref":"Reactor.Template.Input.html#t:t/0"},{"type":"module","title":"Reactor.Template.Result","doc":"The `result` template.","ref":"Reactor.Template.Result.html"},{"type":"type","title":"Reactor.Template.Result.t/0","doc":"","ref":"Reactor.Template.Result.html#t:t/0"},{"type":"module","title":"Reactor.Template.Value","doc":"A statically `value` template.","ref":"Reactor.Template.Value.html"},{"type":"type","title":"Reactor.Template.Value.t/0","doc":"","ref":"Reactor.Template.Value.html#t:t/0"},{"type":"extras","title":"Read Me","doc":"# Reactor\n\n![Elixir CI](https://github.com/ash-project/reactor/actions/workflows/elixir.yml/badge.svg)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Hex version badge](https://img.shields.io/hexpm/v/reactor.svg)](https://hex.pm/packages/reactor)\n\nReactor is a dynamic, concurrent, dependency resolving saga orchestrator.\n\nWoah. That's a lot. Let's break it down:\n\n- **Saga orchestrator** A [saga][saga pattern] is a way of providing\n  transaction-like semantics across multiple distinct resources.\n- **Dependency resolving** reactor allows you to describe the dependencies\n  between your saga steps using _arguments_ which are converted into a\n  [DAG][dag] and used to compute execution order.\n- **Concurrent** unless otherwise specified reactor will run as many steps as\n  possible concurrently whilst taking into account the results of the dependency\n  resolution.\n- **Dynamic** whilst you can define a reactor statically using our awesome DSL,\n  you can also build workflows dynamically - and even add steps while the\n  reactor is running.\n\n[saga pattern](https://learn.microsoft.com/en-us/azure/architecture/reference-architectures/saga/saga)","ref":"readme.html"},{"type":"extras","title":"Sponsors - Read Me","doc":"Thanks to [Alembic Pty Ltd](https://alembic.com.au/) for sponsoring a portion of\nthis project's development.","ref":"readme.html#sponsors"},{"type":"extras","title":"Installation - Read Me","doc":"The package can be installed by adding `reactor` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:reactor, \"~> 0.4.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Documentation - Read Me","doc":"Documentation for the latest release will be [available on\nhexdocs](https://hexdocs.pm/reactor) and for the [`main`\nbranch](https://ash-project.github.io/reactor).","ref":"readme.html#documentation"},{"type":"extras","title":"Contributing - Read Me","doc":"- To contribute updates, fixes or new features please fork and open a\n  pull-request against `main`.\n- Please use [conventional\n  commits](https://www.conventionalcommits.org/en/v1.0.0/) - this allows us to\n  dynamically generate the changelog.\n- Feel free to ask any questions on the `#reactor` channel on the [Ash\n  Discord](https://discord.gg/D7FNG2q).","ref":"readme.html#contributing"},{"type":"extras","title":"Licence - Read Me","doc":"`reactor` is licensed under the terms of the [MIT\nlicense](https://opensource.org/licenses/MIT). See the [`LICENSE` file in this\nrepository](https://github.com/ash-project/reactor/blob/main/LICENSE)\nfor details.\n\n[saga pattern]: https://learn.microsoft.com/en-us/azure/architecture/reference-architectures/saga/saga\n[dag]: https://en.wikipedia.org/wiki/Directed_acyclic_graph","ref":"readme.html#licence"},{"type":"extras","title":"Getting started with Reactor","doc":"# Getting started with Reactor\n\nReactor is a lot of things:\n\n- A [saga orchestrator](https://www.cs.cornell.edu/andru/cs711/2002fa/reading/sagas.pdf).\n- A [composable DSL](https://ash-hq.org/docs/dsl/reactor) for creating workflows.\n- A [builder](https://ash-hq.org/module/reactor/latest/reactor-builder) for dynamically creating workflows.\n- Capable of mixing concurrent and serialised workflows.\n- Resolves dependencies between tasks using a [directed acyclic graph](https://en.wikipedia.org/wiki/Directed_acyclic_graph).\n\nLet's start by breaking down what each of these features are, and how we can use them.","ref":"getting-started-with-reactor.html"},{"type":"extras","title":"Transaction-like semantics with Sagas - Getting started with Reactor","doc":"If you've been building apps long enough you will have used database transactions. They're amazing. Basically, if you put all your changes in a transaction and there's a failure then the transaction gets rolled back and the system is not left in an inconsistent state. This is great when you're working with a single database, but in the modern world we often have to integrate with [many](https://www.postgresql.org/) [different](https://www.mongodb.com/) [kinds](https://redis.io/) [of](https://cassandra.apache.org/_/index.html) [data](https://aws.amazon.com/redshift/) [stores](https://clickhouse.com/) and [as](https://www.salesforce.com/) [many](https://stripe.com/) [different](https://www.twilio.com/en-us) [SaaS](https://www.xero.com/) [products](https://www.vendhq.com/).\n\nOften we need to orchestrate a \"transaction\" across multiple services at once; for example:\n\n1. Register a new user\n2. Create a Stripe customer\n3. Create a Stripe subscription\n4. Send a transactional welcome email via SendGrid\n5. Track the conversion in Salesforce\n\nIf any of these steps fails we may want to retry or roll back depending on what failed and whether it's recoverable. Reactor allows us to do this by defining the `Reactor.Step` behaviour:\n\n```elixir\ndefmodule MyApp.CreateStripeSubscriptionStep do\n  use Reactor.Step\n\n  @impl true\n  def run(arguments, context, options) do\n    Stripe.Subscription.create(arguments.stripe_customer_id, items: [plan: arguments.stripe_plan_id])\n  end\n\n  @impl true\n  def compensate(%{code: :network_error}, arguments, context, options) do\n    :retry\n  end\n\n  def compensate(error, arguments, context, options) do\n    :ok\n  end\n\n  @impl true\n  def undo(subscription, arguments, context, options) do\n    case Stripe.Subscription.delete(subscription) do\n      {:ok, _} -> :ok\n      {:error, reason} -> {:error, reason}\n    end\n  end\nend\n```\n\nHere we've defined a step that tries to create a new Stripe subscription. If that fails then we have defined `compensate/4` check whether it was a network error, and if so ask the Reactor to retry, otherwise we return `:ok` because the subscription wasn't created so we don't have to do any other clean up. Lastly, we define `undo/4` to delete the subscription if the Reactor asks us to undo our work - which it will do if a step later in the workflow fails.\n\nHow do you compose these steps together you ask? Let's discuss that in the next section.","ref":"getting-started-with-reactor.html#transaction-like-semantics-with-sagas"},{"type":"extras","title":"Composing steps with the Reactor DSL - Getting started with Reactor","doc":"Reactor uses [`spark`](https://ash-hq.org/docs/guides/spark/latest/tutorials/get-started-with-spark) to define an easy to use (and easy to extend) DSL for defining workflows. Let's start with an example:\n\n```elixir\ndefmodule MyApp.RegisterUserReactor do\n  use Reactor\n\n  input :email\n  input :password\n  input :plan_name\n\n  step :register_user, MyApp.RegisterUserStep do\n    argument :email, input(:email)\n    argument :password, input(:password)\n  end\n\n  step :create_stripe_customer, MyApp.CreateStripeCustomerStep do\n    argument :email, input(:email)\n  end\n\n  step :find_stripe_plan, MyApp.FindStripePlanStep do\n    argument :plan_name, input(:plan_name)\n  end\n\n  step :create_stripe_subscription, MyApp.CreateStripeSubscriptionStep do\n    argument :customer_id do\n      source result(:create_stripe_customer)\n      transform &(&1.id)\n    end\n\n    argument :plan_id do\n      source result(:get_stripe_plan)\n      transform &(&1.id)\n    end\n  end\n\n  step :send_welcome_email, MyApp.SendWelcomeEmailStep do\n    argument :email, input(:email)\n    argument :_subscription, result(:create_stripe_subscription)\n  end\n\n  step :track_conversion, MyApp.TrackSalesforceConversionStep do\n    argument :email, input(:email)\n    argument :plan_name, input(:plan_name)\n    argument :_welcome_email, result(:send_welcome_email)\n  end\n\n  return :register_user\nend\n```\n\nHere we've defined a Reactor that performs the steps needed for the user registration example above.\n\nHere we define the steps that need to be performed and describe the dependencies between them by way of arguments.\n\n> ##","ref":"getting-started-with-reactor.html#composing-steps-with-the-reactor-dsl"},{"type":"extras","title":"Beware of ordering {: .warning} - Getting started with Reactor","doc":">\n> Whilst the order of the steps in the example makes sense logically it has no\n> effect on the order that the Reactor will execute them in. This is the\n> fundamental difference between Reactor and other Saga tools in the Elixir\n> ecosystem.\n\nThe Reactor will put the steps into a graph with the steps as vertices and the arguments as edges and will find any that have no inbound edges and run as many of them at once as possible. This means that `:register_user`, `:create_stripe_customer` and `:find_stripe_plan` will all execute simultaneously, after which `:create_stripe_subscription`, then `:send_welcome_email` and `:track_conversion` will run. Notice that `:send_welcome_email` and `:track_conversion` have arguments that depend on the results of previews steps that they don't necessarily need, however we want to ensure that they don't run until all the important steps are complete (it's hard to unsend an email!)","ref":"getting-started-with-reactor.html#beware-of-ordering-warning"},{"type":"extras","title":"Building workflows programmatically - Getting started with Reactor","doc":"While using the DSL makes it extremely easy to build a static workflow, sometimes you need the flexibility to build a workflow based on a set of criteria that may not be known at compile time (for example if you're using user input to build the workflow).\n\nEvery Reactor is ultimately an instance of the `Reactor` struct (you can use `Reactor.Info.to_struct(MyApp.RegisterUserReactor)` to see the one for the DSL Reactor above). The functions in `Reactor.Builder` can be used to create an identical workflow:\n\n    iex> reactor = Reactor.Builder.new()\n    ...> {:ok, reactor} = Reactor.Builder.add_input(reactor, :email)\n    ...> {:ok, reactor} = Reactor.Builder.add_input(reactor, :password)\n    ...> {:ok, reactor} = Reactor.Builder.add_input(reactor, :plan_name)\n    ...> {:ok, reactor} = Reactor.Builder.add_step(reactor, :register_user, MyApp.RegisterUserStep, email: {:input, :email}, password: {:input, :password})\n    # etc...","ref":"getting-started-with-reactor.html#building-workflows-programmatically"},{"type":"extras","title":"Mixing concurrent and synchronous workflow steps - Getting started with Reactor","doc":"If you look at the `Reactor.Step` struct, you'll see that it has an `async?` field. This is available as a step DSL option, and as an option to `add_step/5`. It defaults to `true`, however if you set it to `false` Reactor will run the step synchronously. Note that it will only run synchronous steps when it has run out of async steps which can be started. Other than that caveat, Reactor will follow the same ordering rules as specified above.","ref":"getting-started-with-reactor.html#mixing-concurrent-and-synchronous-workflow-steps"},{"type":"extras","title":"Running your Reactor - Getting started with Reactor","doc":"You can pass either a Reactor DSL module or a Reactor struct straight into the `Reactor.run/1..4` function.","ref":"getting-started-with-reactor.html#running-your-reactor"},{"type":"extras","title":"Halting and resuming workflows - Getting started with Reactor","doc":"Any step in the flow can return `{:halt, value}` instead of `{:ok, value}`. When this happens Reactor will wait for any running asynchronous steps to finish and then halt the Reactor returning `{:halted, reactor}`. Later on you can resume the Reactor by passing the halted reactor to the `Reactor.run/1..4` function.","ref":"getting-started-with-reactor.html#halting-and-resuming-workflows"},{"type":"extras","title":"Dynamically adding steps at Reactor run time - Getting started with Reactor","doc":"Any step can return `{:ok, value, additional_steps}` instead of just `{:ok, value}` and those new steps will be added to the graph and have their dependencies calculated automatically. `additional_steps` should be a list of `Reactor.Step` structs, which you can create with `Reactor.Builder.new_step/2..4`.\n\nThere are a couple of additional things to be aware of when steps at run time:\n\n1. If you add steps that induce a cyclic dependency then the Reactor will commence an undo, just as if a step had failed.\n2. Steps can have the same name as other steps and their intermediate results will be replaced.\n3. You are allowed to make dependency cycles in the very specific case of a new\n   step relying on it's own output.","ref":"getting-started-with-reactor.html#dynamically-adding-steps-at-reactor-run-time"}],"content_type":"text/markdown"}